/*****************************************************************************
 * BitStreamReader.cpp: Read from Bitstreams
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

// BitStreamReader.cpp: implementation of the BitStreamReader class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <stdio.h>
#include <string.h>
#include "BitStreamReader.h"

/*
 * --------------------------------------------------------------
	Constructor

	you can change the m_bitbuffer_size to any value.
	set m_bitbuffer_size = 1 for very poor performance :-)
 * --------------------------------------------------------------
 */
BitStreamReader::BitStreamReader()
{
	m_id = NULL;
	m_bitbuffer_size = 2048;
	m_bitbuffer = new unsigned char[m_bitbuffer_size];
	m_bitcounter = 0;
}

/*
 * --------------------------------------------------------------
	Destructor
 * --------------------------------------------------------------
 */
BitStreamReader::~BitStreamReader()
{
	Close();
	if (m_bitbuffer!=NULL)
	{
		delete m_bitbuffer;
		m_bitbuffer = NULL;
	}
}
/*
 * --------------------------------------------------------------
	Open a bitstream

	Previous stream is closed.
 * --------------------------------------------------------------
 */
bool BitStreamReader::Open(char *filename)
{
	Close();
	m_id = fopen(filename,"rb");
	if (m_id!=NULL)
	{
		ResetBitStream();
		return true;
	}
	else
	{
		return false;
	}
}
/*
 * --------------------------------------------------------------
	Read one bit from bitstream
	return (int)1 or (int)0

	m_bitpos start to 7 and decrease to -1
	when m_bitpos is equal to -1 we go to the next byte (m_bytepos)
	of the m_bitbuffer.
	when m_bytepos is at the end of the buffer,
	the buffer is reloaded from disk.

	NOTE: can raise EOFBSException

 * --------------------------------------------------------------
 */
int BitStreamReader::ReadBit()
{
	if (m_bitpos==-1)
	{
		m_bytepos++;
		if (m_bytepos == m_bitbuffer_size)
		{
			LoadBuffer();
		}
		m_bitpos = 7;
	}
	int mask = (1<<m_bitpos);
	int bit = m_bitbuffer[m_bytepos] & mask;
	m_bitpos--;

	// --------------------------------------------
	// update counters...
	//
	// beware of overflow if the bitstream is huge...
	//
	m_bitcounter++; // one bit have been read
	//
	if (m_bitpos==-1)
	{
		m_bytecounter++; // one byte have been read
	}
	// --------------------------------------------


	if (bit!=0)
		return 1;
	else
		return 0;
}
/*
 * --------------------------------------------------------------
	Close the bitstream
 * --------------------------------------------------------------
 */
void BitStreamReader::Close()
{
	if (m_id!=NULL)
	{
		fclose(m_id);
		m_id=NULL;
	}
}

/*
 * --------------------------------------------------------------
	Read some bits from the stream
	count = number of bit to read
	beware: 0 <= count <= 32

	NOTE: can raise EOFBSException

 * --------------------------------------------------------------
 */
unsigned int BitStreamReader::ReadBits(int count)
{
	unsigned int v=0;
	unsigned int bit = 0;
	for (int b=0;b<count;b++)
	{
		bit = ReadBit();
		v = (v<<1) | bit;
	}
	if (bit==-1)
		return -1;
	else
		return v;
}

/*
 * --------------------------------------------------------------
	from ISO/IEC 11172:

	The function nextbits () permits comparison of a bit string
	with the next bits to be decoded in the bit stream.

    strings like "0000 1011 1001" are allowed

	NOTE: can raise EOFBSException

 * --------------------------------------------------------------
 */
bool BitStreamReader::TestNextBits(char *bits)
{
	int v = 0;
	int l = 0;

	v = BitStreamReader::GetValue(bits,l);

	//
	// save the bitstream state
	//
	int prevbytecounter = m_bytecounter;
	int prevbitcounter = m_bitcounter;
	int prevbitpos = m_bitpos;
	int prevbytepos = m_bytepos;
	int next_bits;
	long prevpos = ftell(m_id);

	if (prevpos==0)
	{
		// if no bit have been read
		next_bits = ReadBits(l);
		fseek(m_id,prevpos,SEEK_SET);
	}
	else
	{
		// if a least one bit have been read
		next_bits = ReadBits(l);
		prevpos -= m_bitbuffer_size;
		fseek(m_id,prevpos,SEEK_SET);
		LoadBuffer();
	}

	//
	// restore bistream state
	//

	m_bytecounter = prevbytecounter;
	m_bitcounter = prevbitcounter;
	m_bitpos = prevbitpos;
	m_bytepos = prevbytepos;

	//
	// do the comparison
	//

	if (next_bits == v)
	{
		return true;
	}
	else
	{
		return false;
	}
}

/*
 * --------------------------------------------------------------
	convert a string with bits into integer value
 * --------------------------------------------------------------
 */

int BitStreamReader::GetValue(char *bits)
{
	int l;
	return BitStreamReader::GetValue(bits,l);
}
/*
 * --------------------------------------------------------------
	convert a string with bits into integer value
	return then length in bit of this value in parameter l.

	for example :

	int l;
	bistr.GetValue("1110 1111",l)

	l == 8

 * --------------------------------------------------------------
 */
int BitStreamReader::GetValue(char *bits,int &l)
{
	int v=0;
	int i=0;
	int sl = strlen(bits);
	for (i=0;i<sl;i++)
	{
		if (bits[i] == '0')
		{
			v = (v<<1) | 0;
			l++;
		}
		else if (bits[i]=='1')
		{
			v = (v<<1) | 1;
			l++;
		}
	}
	return v;
}
/*
 * --------------------------------------------------------------
	from ISO/IEC 11172:

The function bytealigned () returns 1 if the current position is
on a byte boundary, that is the next bit in the bitstream is the
first bit in a byte. Otherwise it returns 0.

 * --------------------------------------------------------------
 */
bool BitStreamReader::isByteAligned()
{
	if (m_bitpos==7)
	{
		return true;
	}
	else
	{
		return false;
	}
}

/*
 * --------------------------------------------------------------
	return the byte counter, usefull for framesize calculation
 * --------------------------------------------------------------
 */
unsigned long BitStreamReader::BytePosition()
{
	return m_bytecounter;
}
/*
 * --------------------------------------------------------------
	return the bit counter, usefull for debug
 * --------------------------------------------------------------
 */
unsigned long BitStreamReader::BitPosition()
{
	return m_bitcounter;
}
/*
 * --------------------------------------------------------------
	fill the internal buffer m_bitbuffer with data from disk
	can raise EOFBSException
 * --------------------------------------------------------------
 */

void BitStreamReader::LoadBuffer()
{
	int nb = fread(m_bitbuffer,1,m_bitbuffer_size,m_id);
	if (nb<=0)
	{
		throw EOFBSException();
	}
	else if (nb !=m_bitbuffer_size)
	{
		// last read before EOF
		m_bitbuffer_size = nb;
	}
	m_bytepos = 0;
	m_bitpos = 7;
}

/*
 * --------------------------------------------------------------
	Prepare the bitstream internal state to the first read.
 * --------------------------------------------------------------
 */

void BitStreamReader::ResetBitStream()
{
	m_bitpos = -1;
	m_bytepos = m_bitbuffer_size-1;
	m_bytecounter = 0;
	m_bitcounter = m_bytecounter*8;
}
