/*****************************************************************************
 * BitrateControl.cpp: Bitrate Control Class
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include "StdAfx.h"
#include <math.h>
#include "Transfer.h"
#include "VideoSequence.h"
//#include "bitratecontrol.h"

BitrateControl::BitrateControl(void)
{
	mult_actualized = false;
	quality = 1;
	forceTargetSize = 0;
	sliceSkipProb = 0;
}

BitrateControl::~BitrateControl(void)
{
}

float BitrateControl::QUALITY_QUANTIZER[5] = { 3, 1.8, 1.4, 1.15, 1 };

// set the ratio
void BitrateControl::SetRatio(__int64 actualSizeBytes, __int64 targetSizeBytes)
{
	this->actualSizeBytes = actualSizeBytes;
	this->targetSizeBytes = targetSizeBytes;

	ratio = (float)targetSizeBytes / (float)actualSizeBytes;
	if(ratio > 1) ratio = 1;
	if(ratio < 1) { control_ratio = 1 / ((QUALITY_QUANTIZER[quality] - 1) / (1.6 * ratio) + 1); } //ratio; 
	else          { control_ratio = 1; }

	if(quality == 4) sliceSkipProb = 0;
	else			 sliceSkipProb = 1 - control_ratio;
}

// get compression ratio
float BitrateControl::GetRatio()
{
	return ratio;
}

// get actual compression ratio
float BitrateControl::GetCurrRatio()
{
	return control_ratio;
}

void BitrateControl::SetQuality(int quality)
{
	this->quality = quality;
}

// forces the target size to be 99% (force=1), instead of 95% (0) of the target size
void BitrateControl::ForceTargetSize(int force)
{
	forceTargetSize = force;
}										

// calculate actual compression ratio (multiplier)
float BitrateControl::CalcMultiplier()
{
	float multiplier;
	float security_margin = 0.95;	// margin needed for not to overflow target file size (should be 0.99)
	if(forceTargetSize) security_margin = 0.99; // override security margin

	// if there are enough bytes to make good prediction
	if(outputBytes > 1024 * 1024)
	{
		if(!mult_actualized && ratio < 1)
		{
/*			float seqRatio = ((float)outputBytesSeq) / (float)inputBytesSeq;
			float diff_bytes = (ratio * security_margin / seqRatio - 1) * (float)(targetSizeBytes - outputBytes); // positiv -> berschuss, negativ -> defizit
			float diff_bytesSeq = diff_bytes / outputBytesSeq;
			float newSeqRatio = ((float)outputBytesSeq + diff_bytesSeq) / (float)inputBytesSeq;
			
			float delta_control_ratio = newSeqRatio - seqRatio;*/
			float numFollowSeq = (float)(actualSizeBytes - inputBytes) / (float)inputBytesSeq; // ** estimation **
			float curr_diff_bytes = (float)inputBytes * ratio * security_margin - (float)outputBytes; // ** precise ** positiv -> berschuss, negativ -> defizit
			float curr_diff_bytesSeq = curr_diff_bytes / numFollowSeq; // ** estimation **
			float seqRatio = ((float)outputBytesSeq) / (float)inputBytesSeq; // ** estimation **
			float diff_bytes = (ratio * security_margin / seqRatio - 1) * (float)(targetSizeBytes - outputBytes); // ** estimation ** positiv -> berschuss, negativ -> defizit
			float diff_bytesSeq = diff_bytes / numFollowSeq; // ** estimation **
			float newSeqRatio = ((float)outputBytesSeq + diff_bytesSeq + curr_diff_bytesSeq) / (float)inputBytesSeq; // ** estimation **
			
			float delta_control_ratio = newSeqRatio - seqRatio;

			float target_multiplier = 1 / ((QUALITY_QUANTIZER[quality] - 1) / (1.6 * ratio) + 1);

			// if quantizer should be higher and quantizer is already high
			if(delta_control_ratio < 0 && (control_ratio - target_multiplier) < 0) 
			{
				if(sliceSkipProb < 0.01 || quality == 4)
				{
					control_ratio += 0.1 * delta_control_ratio ;
				}
				else
				{
					sliceSkipProb = 1 - (1 - sliceSkipProb) * 1.04; // sliceSkipProb should get smaller
				}
			}

			// if quantizer should be lower and quantizer is already low
			if(delta_control_ratio > 0 && (control_ratio - target_multiplier) > 0) 
			{
				if(sliceSkipProb >= 1 || quality == 4)
				{
					control_ratio += 0.1 * delta_control_ratio;
				}
				else
				{
					sliceSkipProb = 1 - (1 - sliceSkipProb) / 1.04; // sliceSkipProb should get higher
					if(sliceSkipProb < 0) sliceSkipProb = 0;
				}
			}
			
			// if quantizer should be lower and quantizer is high
			// if quantizer should be higher and quantizer is low
			if(delta_control_ratio >= 0 && (control_ratio - target_multiplier) <= 0 ||
			   delta_control_ratio <= 0 && (control_ratio - target_multiplier) >= 0) 
			{
				control_ratio += 0.1 * delta_control_ratio;// * fabs(control_ratio);
			}

//			control_ratio += 0.1 * (ratio * security_margin - ((float)outputBytes) / (float)inputBytes);
			mult_actualized = true;
		}
		multiplier = control_ratio;
		if(control_ratio < 0.1) control_ratio = 0.1;
	}
	else
	{
		multiplier = control_ratio;//ratio;
	}
	if(multiplier < 0.1) multiplier = 0.1;

	return multiplier;
}

// renews the multiplier
void BitrateControl::UpdateStatistics()
{
	CalcMultiplier();
	mult_actualized = false;
}

// edit a picture
void BitrateControl::ProcessPicture(Picture *Frame)
{
	// change quantizer scales
	qScaleType = Frame->qScaleType;

	// set the slice skipping probability
	Frame->sliceSkipProb = sliceSkipProb * MAX_SLICE_SKIP_PROB;
}

// edit a slice
void BitrateControl::ProcessSlice(Slice *pSlice)
{
	float multiplier = control_ratio;

	// do slice quantizer scale code
	pSlice->quantizerScaleCode = DeEncodeQuantizerScale(qScaleType, pSlice->quantizerScaleCode, multiplier);

	for(int i = 0; i < pSlice->currMB; i++)
	{
		pSlice->MB[i]->quantizerScaleCode = DeEncodeQuantizerScale(qScaleType, pSlice->MB[i]->quantizerScaleCode, multiplier);
	}
}

// edit a slice
/*int BitrateControl::GetMBQuantizerScaleCode(int quantizerScaleCode)
{
	float multiplier = control_ratio;

	// decode and encode new quantizer scale
	return DeEncodeQuantizerScale(qScaleType, quantizerScaleCode, multiplier);
}*/

float BitrateControl::GetSliceSkipProb()
{
	return min(max(sliceSkipProb, 0), 1);
}

int BitrateControl::DeEncodeQuantizerScale(int qScaleType, int quantizerScaleCode, float multiplier)
{
	int testQuantizerScaleCode = quantizerScaleCode;

	int quantizerScale = DecodeQuantizerScale(qScaleType, quantizerScaleCode);
	// quantizer scale my only grow or stay the same but never get lower
	if(multiplier < 1) 
	{
		quantizerScale = floor(quantizerScale / multiplier);
	}
	if(quantizerScale == 0) quantizerScale = 1;

    quantizerScaleCode = EncodeQuantizerScale(qScaleType, quantizerScale);

	// code may never get lower
	_ASSERT(quantizerScaleCode >= testQuantizerScaleCode);

	return quantizerScaleCode;
}


// Quantizer Scale Map (derived from quantizer_scale_code and q_scale_type) - ISO 13818-2 Table 7-6 (q_scale_type = 0)
const unsigned char BitrateControl::QuantizerScaleMap0[31] =
{
	2,
	4,
	6,
	8,
	10,
	12,
	14,
	16,
	18,
	20,
	22,
	24,
	26,
	28,
	30,
	32,
	34,
	36,
	38,
	40,
	42,
	44,
	46,
	48,
	50,
	52,
	54,
	56,
	58,
	60,
	62
};

// Inverse Quantizer Scale Map (derived from quantizer_scale_code and q_scale_type) - ISO 13818-2 Table 7-6 (q_scale_type = 0)
const unsigned char BitrateControl::QuantizerScaleMap0INV[62] =
{
	0,
	0,
	1,
	1,
	2,
	2,
	3,
	3,
	4,
	4,
	5,
	5,
	6,
	6,
	7,
	7,
	8,
	8,
	9,
	9,
	10,
	10,
	11,
	11,
	12,
	12,
	13,
	13,
	14,
	14,
	15,
	15,
	16,
	16,
	17,
	17,
	18,
	18,
	19,
	19,
	20,
	20,
	21,
	21,
	22,
	22,
	23,
	23,
	24,
	24,
	25,
	25,
	26,
	26,
	27,
	27,
	28,
	28,
	29,
	29,
	30,
	30,
};

// Quantizer Scale Map (derived from quantizer_scale_code and q_scale_type) - ISO 13818-2 Table 7-6 (q_scale_type = 1)
const unsigned char BitrateControl::QuantizerScaleMap1[31] =
{
	1,
	2,
	3,
	4,
	5,
	6,
	7,
	8,
	10,
	12,
	14,
	16,
	18,
	20,
	22,
	24,
	28,
	32,
	36,
	40,
	44,
	48,
	52,
	56,
	64,
	72,
	80,
	88,
	96,
	104,
	112
};

// Inverse Quantizer Scale Map (derived from quantizer_scale_code and q_scale_type) - ISO 13818-2 Table 7-6 (q_scale_type = 1)
const unsigned char BitrateControl::QuantizerScaleMap1INV[112] =
{
	0,
	1,
	2,
	3,
	4,
	5,
	6,
	7,
	8,
	8,
	9,
	9,
	10,
	10,
	11,
	11,
	12,
	12,
	13,
	13,
	14,
	14,
	15,
	15,
	15,
	15,
	16,
	16,
	16,
	16,
	17,
	17,
	17,
	17,
	18,
	18,
	18,
	18,
	19,
	19,
	19,
	19,
	20,
	20,
	20,
	20,
	21,
	21,
	21,
	21,
	22,
	22,
	22,
	22,
	23,
	23,
	23,
	23,
	23,
	23,
	24,
	24,
	24,
	24,
	24,
	24,
	24,
	24,
	25,
	25,
	25,
	25,
	25,
	25,
	25,
	25,
	26,
	26,
	26,
	26,
	26,
	26,
	26,
	27,
	27,
	27,
	27,
	27,
	27,
	27,
	27,
	28,
	28,
	28,
	28,
	28,
	28,
	28,
	28,
	29,
	29,
	29,
	29,
	29,
	29,
	29,
	29,
	30,
	30,
	30,
	30,
	30
};

// decode a quantizer scale code to a quantizer scale using the quantizer scale maps
int BitrateControl::DecodeQuantizerScale(int qScaleType, int quantizerScaleCode)
{
	int quantizer_scale;

	_ASSERT(quantizerScaleCode < 32);
	_ASSERT(quantizerScaleCode > 0);

	// determine which quantizer scale to use
	if(qScaleType) quantizer_scale = QuantizerScaleMap1[quantizerScaleCode - 1];
	else		   quantizer_scale = QuantizerScaleMap0[quantizerScaleCode - 1];

	return quantizer_scale;
}

// encode a quantizer scale to a quantizer scale code using the quantizer scale maps
int BitrateControl::EncodeQuantizerScale(int qScaleType, int quantizerScale)
{
	int quantizer_scale_code;

	// determine which quantizer scale to use
	if(qScaleType) 
	{
		if(quantizerScale < 1) quantizerScale = 1;
		if(quantizerScale > 112) quantizerScale = 112;
		quantizer_scale_code = QuantizerScaleMap1INV[quantizerScale - 1] + 1;
	}
	else
	{
		if(quantizerScale < 2) quantizerScale = 2;
		if(quantizerScale > 62) quantizerScale = 62;
		quantizer_scale_code = QuantizerScaleMap0INV[quantizerScale - 2] + 1;
	}

	return quantizer_scale_code;
}

// resets the multiplier actualized flag so next time CalcMultiplier is called the multiplier ist calculated again
float BitrateControl::RecalcRatio()
{
	mult_actualized = false;
	return CalcMultiplier();
}
