/*****************************************************************************
 * IFO.cpp: Analyse IFO Class
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include "StdAfx.h"
#include "ifo.h"

IFO::IFO(void)
{
	movieOnly = 0;

	isVMG = 0;

	IFOBufferSize = 0;
	currIFOBufferSize = 0;
	
	markSECT = NULL;
	AVStreamID = NULL;

	sectorLUT = new SECTOR_LUT[1024 * 1024];
}

IFO::~IFO(void)
{
	if(IFOBufferSize > 0) delete IFOBuffer;
	if(markSECT != NULL) delete markSECT;
	if(AVStreamID != NULL) delete AVStreamID;

	delete sectorLUT;
}

#define IFO_READ  false
#define IFO_WRITE true

// reads original IFO/BUP and creates the sector lookup table
int IFO::Load(char *filePath)
{

	#ifdef _DEBUG
	for(int i = 0; i < 1024 * 1024; i++)
	{
		sectorLUT[i].srcSector    = -1;
		sectorLUT[i].targetSector = -1;
	}
	for(i = 0; i < 128; i++)
	{
		titleStartLUT[i].srcSector = -1;
		titleStartLUT[i].targetSector = -1;
	}
	#endif

	// open file
	HANDLE hFile = CreateFile(filePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);

	_ASSERT(hFile > 0);

	// get file size
	int fileSize, readSize;
	fileSize = GetFileSize(hFile, NULL);

	// realloc buffer if too small
	if(IFOBufferSize < fileSize)
	{
		if(IFOBufferSize != 0) delete IFOBuffer;
		IFOBuffer = new unsigned char[fileSize];
		IFOBufferSize = fileSize;
	}
	currIFOBufferSize = fileSize;

	// read whole file
	ReadFile(hFile, IFOBuffer, fileSize, (LPDWORD)&readSize, NULL);

	_ASSERT(readSize == fileSize);

	// close file
	CloseHandle(hFile);

	// check if VMG or VTS
	isVMG = -1;
	if(strncmp((const char *)IFOBuffer, "DVDVIDEO-VMG", 12) == 0) isVMG = 1;
	if(strncmp((const char *)IFOBuffer, "DVDVIDEO-VTS", 12) == 0) isVMG = 0;
	_ASSERT(isVMG != -1);

	// process structure (read, create sector LUTs)
	if(isVMG)
	{
		ProcessVMGSectorLUTs(IFO_READ);
	}
	else
 	{
		ProcessVTSSectorLUTs(IFO_READ);
	}

	return TRUE;
}

// writes modified IFO/BUP from the sector lookup table
int IFO::Save(char *filePath)
{
	// process structure (write, interpret sector LUTs)
	if(isVMG)
	{
		// if <not> in movie only mode - process the VIDEO_TS.IFO
		if(!movieOnly) ProcessVMGSectorLUTs(IFO_WRITE);
	}
	else
	{
		ProcessVTSSectorLUTs(IFO_WRITE);
	}

	// create BUP name and path	
	char bupFilePath[1024];
	strcpy(bupFilePath, filePath);
	strcpy(&bupFilePath[strlen(bupFilePath) - 3], "BUP");

	// open file
	HANDLE hFile = CreateFile(filePath, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	_ASSERT(hFile > 0);
	int writeSize;
	// write whole file
	WriteFile(hFile, IFOBuffer, currIFOBufferSize, (LPDWORD)&writeSize, NULL);
	_ASSERT(writeSize == currIFOBufferSize);
	// close file
	CloseHandle(hFile);

	// open file
	hFile = CreateFile(bupFilePath, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	_ASSERT(hFile > 0);
	// write whole file
	WriteFile(hFile, IFOBuffer, currIFOBufferSize, (LPDWORD)&writeSize, NULL);
	_ASSERT(writeSize == currIFOBufferSize);
	// close file
	CloseHandle(hFile);

	return TRUE;
}

// close file and leave unmodified
int IFO::Close()
{
	isVMG = -1;
	return TRUE;
}

#define intCastX(z, x)  ((IFOBuffer[x]     << 24) + \
                         (z[x + 1] << 16) + \
				         (z[x + 2] <<  8) + \
					      z[x + 3])

#define sintCastX(z, x) ((z[x]     <<  8) + \
					      z[x + 1])

#define byteCastX(z, x) z[x]

#define intCast(x)  ((IFOBuffer[x]     << 24) + \
                     (IFOBuffer[x + 1] << 16) + \
				     (IFOBuffer[x + 2] <<  8) + \
					  IFOBuffer[x + 3])

#define sintCast(x) ((IFOBuffer[x]     <<  8) + \
					  IFOBuffer[x + 1])

#define byteCast(x) IFOBuffer[x]

#define intCastW(x, y)  { IFOBuffer[x]     = (y >> 24) & 0xFF; \
						  IFOBuffer[x + 1] = (y >> 16) & 0xFF; \
						  IFOBuffer[x + 2] = (y >>  8) & 0xFF; \
						  IFOBuffer[x + 3] = y; }

#define sintCastW(x, y) { IFOBuffer[x]     = (y >>  8) & 0xFF; \
						  IFOBuffer[x + 1] = y; }

#define byteCastW(x, y) { IFOBuffer[x] = y; }

// processes the IFO structure and creates the sector lookup tables (for VIDEO_TS.IFO)
void IFO::ProcessVMGSectorLUTs(bool write)
{
	int i;
	int LUT = 0, titleLUT = 0;
	SECTOR_LUT *pLUT = sectorLUT;

	if(write) intCastW(0x000C, lastBUPSector)
	else lastBUPSector = intCast(0x000C);

	int IFOLastSector = intCast(0x001C);

	int pTT_SRPT = intCast(0x00C4) * 2048;
	int pVMGM_PGCI_UT = intCast(0x00C8) * 2048;
	int pVMGM_C_ADT = intCast(0x00D8) * 2048;
	int pVMGM_VOBU_ADMAP = intCast(0x00DC) * 2048;

	// for movie only mode - delete all menu related attributes
	if(movieOnly)
	{
		int VMG_VTS_ATRT = intCast(0x00D0) * 2048;
		int numVTS = sintCast(VMG_VTS_ATRT);
//		int lastByteVTS_ATRT = intCast(VMG_VTS_ATRT + 4);
		for(int i = 0; i < numVTS; i++)
		{
			int VTS_ATRT1 = VMG_VTS_ATRT + intCast(VMG_VTS_ATRT + i * 4 + 8);

			// menu audio and subpicture attributes
			sintCastW(VTS_ATRT1 + 8 + 0x0000, 0);
			sintCastW(VTS_ATRT1 + 8 + 0x0002, 0);
			ZeroMemory(&IFOBuffer[VTS_ATRT1 + 8 + 0x0004], 8 * 8);
			sintCastW(VTS_ATRT1 + 8 + 0x0054, 0);
			ZeroMemory(&IFOBuffer[VTS_ATRT1 + 8 + 0x0056], 28 * 6);
		}
	}

/*	// set to region free
	if(write)
	{
		byteCastW(0x0023, 0x00);
	}*/

	// process menu start sector
	if(write) 
	{
		//_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);
/*		_ASSERT(pLUT[LUT].targetSector >= 0);
		_ASSERT(intCast(0x00C0) == pLUT[LUT].srcSector);
		intCastW(0x00C0, pLUT[LUT].targetSector); LUT++;*/
		intCastW(0x00C0, MenuVOBStartSECT);
/*		_ASSERT(titleStartLUT[titleLUT].targetSector >= 0);
		_ASSERT(intCast(0x00C0) == titleStartLUT[titleLUT].srcSector)
		intCastW(0x00C0, titleStartLUT[titleLUT].targetSector); titleLUT++;*/
	}
	else
	{
//		titleStartLUT[titleLUT++].srcSector = intCast(0x00C0);
		MenuVOBStartSECT = intCast(0x00C0);
	}

	// process TT_SRPT (title set beginnings)
	if(pTT_SRPT != 0)
	{
		int numTitles = sintCast(pTT_SRPT);
		_ASSERT(numTitles < 100);
		for(i = 0; i < numTitles; i++)
		{
			if(write) 
			{
				//_ASSERT(titleStartLUT[titleLUT].targetSector == titleStartLUT[titleLUT].srcSector);
				// set title beginning only if 
				_ASSERT(titleStartLUT[titleLUT].targetSector >= 0);
				_ASSERT(intCast(pTT_SRPT + (i * 12 + 8) + 8) == titleStartLUT[titleLUT].srcSector);
				intCastW(pTT_SRPT + (i * 12 + 8) + 8, titleStartLUT[titleLUT].targetSector); 
				titleLUT++;
			}
			else
			{
				titleStartLUT[titleLUT].reference  = REFERENCE_IFO;
				titleStartLUT[titleLUT].srcSector  = intCast(pTT_SRPT + (i * 12 + 8) + 8);
				titleStartLUT[titleLUT++].userData = byteCast(pTT_SRPT + (i * 12 + 8) + 6); // Video Title Set Number (VTSN)
			}
		}
	}
	titleStartLUTEntries = titleLUT;

	if(pVMGM_PGCI_UT != 0)    LUT = ProcessPGCI_UT(pVMGM_PGCI_UT, LUT, write);			// process VMGM_PGCI_UT (program chains)

	if(pVMGM_C_ADT != 0)	  LUT = ProcessC_ADT(pVMGM_C_ADT, LUT, MENU_VOB, write);				// process VMGM_C_ADT (pointer to cells)
	if(pVMGM_VOBU_ADMAP != 0) LUT = ProcessVOBU_ADMAP(pVMGM_VOBU_ADMAP, LUT, MENU_VOB, write);	// process VMGM_VOBU_ADMAP (pointer to VOBUs)

	sectorLUTEntries = LUT;
	currLUTVOBEntry = 0;
	currLUTIFOEntry = 0;

/*	// ignore VOB sectors which point in the IFO
	if(!write)
	{
		for(i = 0; i < sectorLUTEntries; i++)
		{
			if(pLUT[i].srcSector < IFOLastSector) pLUT[i].targetSector = pLUT[i].srcSector;
		}
	}*/

}

// processes the IFO structure and creates the sector lookup tables (for VTS_xx_x.IFO)
void IFO::ProcessVTSSectorLUTs(bool write)
{
	int i;
	int LUT = 0, titleLUT = 0;
	SECTOR_LUT *pLUT = sectorLUT;

	// for movie only mode - delete menu related tables and attributes
	//						 delete menu VOB start sector
	if(movieOnly)
	{
		MenuVOBStartSECT = 0;

		// menu tables
		intCastW(0x00D0, 0);
		intCastW(0x00D8, 0);
		intCastW(0x00DC, 0);
		// menu audio and subpicture attributes
		sintCastW(0x0100, 0);
		sintCastW(0x0102, 0);
		ZeroMemory(&IFOBuffer[0x0104], 8 * 8);
		sintCastW(0x0154, 0);
		ZeroMemory(&IFOBuffer[0x0156], 28 * 6);
	}

	if(write) intCastW(0x000C, lastBUPSector)
	else 	  lastBUPSector = intCast(0x000C);

	int IFOLastSector = intCast(0x001C);

	int pVTS_PGCI = intCast(0x0CC) * 2048;
	int pVTSM_PGCI_UT = intCast(0x0D0) * 2048;
	int pVTS_TMAPTI = intCast(0x0D4) * 2048;
	int pVTSM_C_ADT = intCast(0x0D8) * 2048;
	int pVTSM_VOBU_ADMAP = intCast(0x0DC) * 2048;
	int pVTS_C_ADT = intCast(0x0E0) * 2048;
	int pVTS_VOBU_ADMAP = intCast(0x0E4) * 2048;

	// process menu and title start sector
	if(write) 
	{
		if(intCast(0x00C0) != 0) 
		{
/*			//_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);
			_ASSERT(pLUT[LUT].targetSector >= 0);
			_ASSERT(intCast(0x00C0) == pLUT[LUT].srcSector); intCastW(0x00C0, pLUT[LUT].targetSector); LUT++;*/
			intCastW(0x00C0, MenuVOBStartSECT);
		}
		if(intCast(0x00C4) != 0) 
		{
/*			_ASSERT(pLUT[LUT].targetSector >= 0);
			//_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);
			_ASSERT(intCast(0x00C4) == pLUT[LUT].srcSector); intCastW(0x00C4, pLUT[LUT].targetSector); LUT++;*/
			intCastW(0x00C4, TitleVOBStartSECT);
		}
/*		_ASSERT(titleStartLUT[titleLUT    ].targetSector >= 0);
		_ASSERT(titleStartLUT[titleLUT + 1].targetSector >= 0);
		_ASSERT(intCast(0x00C0) == titleStartLUT[titleLUT].srcSector); intCastW(0x00C0, titleStartLUT[titleLUT].targetSector); titleLUT++;
		_ASSERT(intCast(0x00C4) == titleStartLUT[titleLUT].srcSector); intCastW(0x00C4, titleStartLUT[titleLUT].targetSector); titleLUT++;*/
	}
	else
	{
/*		titleStartLUT[titleLUT++].srcSector = intCast(0x00C0);
		titleStartLUT[titleLUT++].srcSector = intCast(0x00C4);*/
/*		if(intCast(0x00C0) != 0) { pLUT[LUT].reference = REFERENCE_IFO; pLUT[LUT++].srcSector = intCast(0x00C0); }
		if(intCast(0x00C4) != 0) { pLUT[LUT].reference = REFERENCE_IFO; pLUT[LUT++].srcSector = intCast(0x00C4); }*/
		if(intCast(0x00C0) != 0) {  MenuVOBStartSECT = intCast(0x00C0); }
		if(intCast(0x00C4) != 0) { TitleVOBStartSECT = intCast(0x00C4); }
	}

/*	titleStartLUTEntries = titleLUT;
	currTitleStartLUTEntry = 0;*/

	if(pVTS_PGCI != 0)						LUT = ProcessPGCI_LU(pVTS_PGCI, LUT, TITLE_VOB, write);	// process VTS_PGCI (program chains)
	if(pVTSM_PGCI_UT != 0 && !movieOnly)    LUT = ProcessPGCI_UT(pVTSM_PGCI_UT, LUT, write);		// process VTSM_PGCI_UT (program chains)

	// read VTS_TMAPTI (pointer to cells)
	if(pVTS_TMAPTI != 0)
	{
		int numProgramChains = sintCast(pVTS_TMAPTI);
		int lastByteVTS_TMAPTI = intCast(pVTS_TMAPTI + 4);// + pVTS_TMAPTI;
		for(int j = 0; j < numProgramChains; j++)
		{
			int poffVTS_TMAP = pVTS_TMAPTI + (j * 4) + 8;
			// check whether language pointer is beyond VMGM_LU border
			_ASSERT(poffVTS_TMAP - pVTS_TMAPTI/**/ < lastByteVTS_TMAPTI);
			int pVTS_TMAP = intCast(poffVTS_TMAP) + pVTS_TMAPTI;

			int	numEntries = sintCast(pVTS_TMAP + 2);
			// read all time entries
			for(int k = 0; k < numEntries; k++)
			{
				if(write) 
				{
					//_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);
					_ASSERT(pLUT[LUT].targetSector >= 0);
					_ASSERT((intCast(pVTS_TMAP + k * 4 + 4) & 0x7FFFFFFF) == pLUT[LUT].srcSector);
					intCastW((pVTS_TMAP + k * 4 + 4), ((intCast(pVTS_TMAP + k * 4 + 4) & 0x80000000) + pLUT[LUT].targetSector)); LUT++;
				}
				else
				{
					pLUT[LUT].reference = REFERENCE_TITLE_VOB;
					pLUT[LUT++].srcSector = intCast(pVTS_TMAP + k * 4 + 4) & 0x7FFFFFFF;
				}
			}
		}
	}

	if(pVTSM_C_ADT != 0 && !movieOnly)		LUT = ProcessC_ADT(pVTSM_C_ADT, LUT, MENU_VOB, write);				// process VTSM_C_ADT (pointer to cells)
	if(pVTSM_VOBU_ADMAP != 0 && !movieOnly) LUT = ProcessVOBU_ADMAP(pVTSM_VOBU_ADMAP, LUT, MENU_VOB, write);	// process VTSM_VOBU_ADMAP (pointer to VOBUs)

	if(pVTS_C_ADT != 0)		  LUT = ProcessC_ADT(pVTS_C_ADT, LUT, TITLE_VOB, write);					// process VTS_C_ADT (pointer to cells)
	if(pVTS_VOBU_ADMAP != 0)  LUT = ProcessVOBU_ADMAP(pVTS_VOBU_ADMAP, LUT, TITLE_VOB, write);		// process VTS_VOBU_ADMAP (pointer to VOBUs)

	sectorLUTEntries = LUT;
	currLUTVOBEntry = 0;
	currLUTIFOEntry = 0;

/*	// ignore VOB sectors which point in the IFO
	if(!write)
	{
		for(i = 0; i < sectorLUTEntries; i++)
		{
			if(pLUT[i].srcSector < IFOLastSector) pLUT[i].targetSector = pLUT[i].srcSector;
		}
	}*/
}

// processes C_ADT structures
int IFO::ProcessC_ADT(int pC_ADT, int LUT, int isMenu, bool write)
{
	SECTOR_LUT *pLUT = sectorLUT;
	int reference = isMenu ? REFERENCE_MENU_VOB : REFERENCE_TITLE_VOB;

	// process C_ADT structure (pointer to cells)
	int lastByteC_ADT = intCast(pC_ADT + 4);// + pC_ADT;
	int pEntryC_ADT = pC_ADT + 8;
	while(/*pEntryC_ADT + 12*/pEntryC_ADT + 11 - pC_ADT <= lastByteC_ADT)
	{
		if(write)
		{
			//_ASSERT(pLUT[LUT    ].targetSector == pLUT[LUT    ].srcSector);
			//_ASSERT(pLUT[LUT + 1].targetSector == pLUT[LUT + 1].srcSector);
			_ASSERT(pLUT[LUT    ].targetSector >= 0);
			_ASSERT(pLUT[LUT + 1].targetSector >= 0);
			_ASSERT(intCast(pEntryC_ADT + 4) == pLUT[LUT].srcSector); intCastW(pEntryC_ADT + 4, pLUT[LUT].targetSector); LUT++;
			_ASSERT(intCast(pEntryC_ADT + 8) == pLUT[LUT].srcSector); intCastW(pEntryC_ADT + 8, pLUT[LUT].targetSector); LUT++;	
		}
		else
		{
			pLUT[LUT].reference = reference; pLUT[LUT++].srcSector = intCast(pEntryC_ADT + 4);	
			pLUT[LUT].reference = reference; pLUT[LUT++].srcSector = intCast(pEntryC_ADT + 8);	
		}
		pEntryC_ADT += 12;
	}

	return LUT;
}

// processes VOBU_ADMAP structures
int IFO::ProcessVOBU_ADMAP(int pVOBU_ADMAP, int LUT, int isMenu, bool write)
{
	SECTOR_LUT *pLUT = sectorLUT;
	int reference = isMenu ? REFERENCE_MENU_VOB : REFERENCE_TITLE_VOB;

	// process VOBU_ADMAP structure (pointer to VOBUs)
	int lastByteVOBU_ADMAP = intCast(pVOBU_ADMAP);// + pVOBU_ADMAP;
	int pEntryVOBU_ADMAP = pVOBU_ADMAP + 4;
	while(pEntryVOBU_ADMAP + 3 - pVOBU_ADMAP <= lastByteVOBU_ADMAP)
	{
		if(write) 
		{
			if(intCast(pEntryVOBU_ADMAP) != 0) 
			{
				//_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);
				_ASSERT(pLUT[LUT].targetSector >= 0);
				_ASSERT(intCast(pEntryVOBU_ADMAP) == pLUT[LUT].srcSector); 
				intCastW(pEntryVOBU_ADMAP, pLUT[LUT].targetSector); LUT++; 
			}
		}
		else 
		{
			pLUT[LUT].reference = reference; 
			if(intCast(pEntryVOBU_ADMAP) != 0) pLUT[LUT++].srcSector = intCast(pEntryVOBU_ADMAP);	
		}
		pEntryVOBU_ADMAP += 4;
	}

	return LUT;
}

// processes PGCI_UT structures
int IFO::ProcessPGCI_UT(int pPGCI_UT, int LUT, bool write)
{
	SECTOR_LUT *pLUT = sectorLUT;

	// process VMGM_PGCI_UT (program chains)
	int numLanguageUnits = sintCast(pPGCI_UT);
	int lastBytePGCI_UT = intCast(pPGCI_UT + 4) + pPGCI_UT;
	// process all Language Units
	for(int i = 0; i < numLanguageUnits; i++)
	{
		int poffPGCI_LU = pPGCI_UT + (i * 8 + 4) + 8;
		// check whether language pointer is beyond VMGM_PGCI_UT border
		_ASSERT(poffPGCI_LU < lastBytePGCI_UT);
		int pPGCI_LU = intCast(poffPGCI_LU) + pPGCI_UT;

		// process all Program Chains
		LUT = ProcessPGCI_LU(pPGCI_LU, LUT, MENU_VOB, write);
	}

	return LUT;
}

// processes PGCI/LU structures
int IFO::ProcessPGCI_LU(int pPGCI_LU, int LUT, int isMenu, bool write)
{
	SECTOR_LUT *pLUT = sectorLUT;
	int reference = isMenu ? REFERENCE_MENU_VOB : REFERENCE_TITLE_VOB;

	int numProgramChains = sintCast(pPGCI_LU);
	int lastBytePGCI_LU = intCast(pPGCI_LU + 4) + pPGCI_LU;
	// process all Program Chains
	for(int j = 0; j < numProgramChains; j++)
	{
		int poffPGC = pPGCI_LU + (j * 8 + 4) + 8;
		// check whether language pointer is beyond VMGM_LU border
		_ASSERT(poffPGC < lastBytePGCI_LU);
		int	pPGC = intCast(poffPGC) + pPGCI_LU;
			
		int numCells = byteCast(pPGC + 0x0003);
		int	pcellInfo = sintCast(pPGC + 0x00E8) + pPGC;
		// process all Cells
		for(int k = 0; k < numCells; k++)
		{
			int pCell = pcellInfo + k * 0x0018;
			if(write)
			{
				if(intCast(pCell + 0x0008) != 0) { /*_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);*/ _ASSERT(pLUT[LUT].targetSector >= 0); _ASSERT(intCast(pCell + 0x0008) == pLUT[LUT].srcSector); intCastW(pCell + 0x0008, pLUT[LUT].targetSector); LUT++; }
				if(intCast(pCell + 0x000C) != 0) { /*_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);*/ _ASSERT(pLUT[LUT].targetSector >= 0); _ASSERT(intCast(pCell + 0x000C) == pLUT[LUT].srcSector); intCastW(pCell + 0x000C, pLUT[LUT].targetSector); LUT++; }
				if(intCast(pCell + 0x0010) != 0) { /*_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);*/ _ASSERT(pLUT[LUT].targetSector >= 0); _ASSERT(intCast(pCell + 0x0010) == pLUT[LUT].srcSector); intCastW(pCell + 0x0010, pLUT[LUT].targetSector); LUT++; }
				if(intCast(pCell + 0x0014) != 0) { /*_ASSERT(pLUT[LUT].targetSector == pLUT[LUT].srcSector);*/ _ASSERT(pLUT[LUT].targetSector >= 0); _ASSERT(intCast(pCell + 0x0014) == pLUT[LUT].srcSector); intCastW(pCell + 0x0014, pLUT[LUT].targetSector); LUT++; }
			}
			else
			{
				pLUT[LUT].reference = reference; if(intCast(pCell + 0x0008) != 0) pLUT[LUT++].srcSector = intCast(pCell + 0x0008);
				pLUT[LUT].reference = reference; if(intCast(pCell + 0x000C) != 0) pLUT[LUT++].srcSector = intCast(pCell + 0x000C);
				pLUT[LUT].reference = reference; if(intCast(pCell + 0x0010) != 0) pLUT[LUT++].srcSector = intCast(pCell + 0x0010);
				pLUT[LUT].reference = reference; if(intCast(pCell + 0x0014) != 0) pLUT[LUT++].srcSector = intCast(pCell + 0x0014);
			}
		}			
	}

	return LUT;
}

// manipulation routines

// selects the movie title (using some heuristics)
int IFO::SearchMovieTitle(int *PGCidx)
{
	int bestMatchTT = -1;
	int maxChapters = -1;
	*PGCidx = -1;

	_ASSERT(!isVMG);
	if(movieOnly)
	{
		int pPGCI = intCast(0x0CC) * 2048;

		int numProgramChains = sintCast(pPGCI);
		int lastBytePGCI = intCast(pPGCI + 4) + pPGCI;

		// process all Program Chains
		for(int j = 0; j < numProgramChains; j++)
		{
			int poffPGC = pPGCI + (j * 8 + 4) + 8;
			int TT = byteCast(pPGCI + (j * 8) + 8) & 0x7F;
			// check whether language pointer is beyond VMGM_LU border
			_ASSERT(poffPGC < lastBytePGCI);
			int	pPGC = intCast(poffPGC) + pPGCI;
				
			int numPGN = byteCast(pPGC + 0x0002);

			// see which title has the most chapters
			if(maxChapters < numPGN)
			{
				bestMatchTT = TT;
				maxChapters = numPGN;
				*PGCidx = j;
			}
		}
	}

	return bestMatchTT;
}

// makes movie title playable (selects proper language and subtitle, makes PGC an entry one, removes prohibited user ops)
void IFO::MakeMovieTitlePlayable(int TT, int PGCidx, int defAud, int defSub, char *selAudFlags, char *selSubFlags)
{
	_ASSERT(!isVMG);
	if(movieOnly)
	{
		int pPGCI = intCast(0x0CC) * 2048;

		int numProgramChains = sintCast(pPGCI);
		int lastBytePGCI = intCast(pPGCI + 4) + pPGCI;

		// process all Program Chains
		for(int j = 0; j < numProgramChains; j++)
		{
			int poffPGC = pPGCI + (j * 8 + 4) + 8;
			int TTidx = byteCast(pPGCI + (j * 8) + 8) & 0x7F;
			// check whether language pointer is beyond VMGM_LU border
			_ASSERT(poffPGC < lastBytePGCI);
			int	pPGC = intCast(poffPGC) + pPGCI;
				
			int numPGN = byteCast(pPGC + 0x0002);

			// if title and pgc is the choosen one
			if(TTidx == TT && PGCidx == j)
			{
				// make PGC entry PGC of TT
				byteCastW(pPGCI + (j * 8) + 8, ((1 << 7) | TTidx));
				// remove prohibited user ops
				intCastW(pPGC + 0x0008, 0);
				// select audio flags
				for(int s = 0; s < 8; s++)
				{
					int val = byteCast(pPGC + 0x000C + s * 2);
					if(selAudFlags[s]/* && s < GetNumLang(REFERENCE_TITLE_VOB)*/) val |= 0x80;
					else													  val &= 0x7F;
					byteCastW(pPGC + 0x000C + s * 2, val);
				}
				// select subpicture flags
				for(s = 0; s < 32; s++)
				{
					int val = byteCast(pPGC + 0x001C + s * 4);
					if(selSubFlags[s]/* && s < GetNumSub(REFERENCE_TITLE_VOB)*/) val |= 0x80;
					else													 val &= 0x7F;
					byteCastW(pPGC + 0x001C + s * 4, val);
				}
				// delete PGC commands
				int pCMD = sintCast(pPGC + 0x00E4) + pPGC;
				int numCmd = sintCast(pCMD + 0x0000) + sintCast(pCMD + 0x0002);
				sintCastW(pCMD + 0x0000, numCmd);
				sintCastW(pCMD + 0x0002, 0);
				ZeroMemory(&IFOBuffer[pCMD + 0x0008], numCmd * 8);
				// add standard audio and subtitle settings
				if(numCmd > 0)
				{
					byteCastW(pCMD + 0x0008 + 0, 0x51);
					byteCastW(pCMD + 0x0008 + 1, 0x00);
					byteCastW(pCMD + 0x0008 + 2, 0x00);
					if(defAud != -1) byteCastW(pCMD + 0x0008 + 3, 0x80 | defAud)
					else			 byteCastW(pCMD + 0x0008 + 3, 0x80 | 0x0F);	// kein audio ausgewhlt
					if(defSub != -1) byteCastW(pCMD + 0x0008 + 4, 0x80 | defSub)
					else			 byteCastW(pCMD + 0x0008 + 4, 0x80 | 0x3E); // kein subpicture ausgewhlt
					byteCastW(pCMD + 0x0008 + 5, 0x00); // 0x80 | stdAngle
					byteCastW(pCMD + 0x0008 + 6, 0x00);
					byteCastW(pCMD + 0x0008 + 7, 0x00);
				}
			}
			else if(TTidx == TT)
			{
				// delete entry status of PGC from TT
				byteCastW(pPGCI + (j * 8) + 8, ((0 << 7) | TTidx));				
			}

		}
	}
}

// selects the movie title (using some heuristics)
void IFO::SelectMovieTitle(int TT, int VTS, unsigned char *orgIFOBuffer)
{
	_ASSERT(isVMG);
	if(movieOnly)
	{
		int pTT_SRPT = intCast(0x0C4) * 2048;

		int porgTT_SRPT = intCastX(orgIFOBuffer, 0x0C4) * 2048;
		int numorgTitles = sintCastX(orgIFOBuffer, porgTT_SRPT);
		int lastByteorgTT_SRPT = intCastX(orgIFOBuffer, porgTT_SRPT + 4) + porgTT_SRPT;

		// process all Program Chains
		for(int j = 0; j < numorgTitles; j++)
		{
			int pTT = porgTT_SRPT + (j * 12) + 8;
			// check whether pointer is beyond porgTT_SRPT border
			_ASSERT(pTT < lastByteorgTT_SRPT);

			int iVTS = byteCastX(orgIFOBuffer, (pTT + 6));
			int iTT  = byteCastX(orgIFOBuffer, (pTT + 7));
			
			// if title set found in original VIDEO_TS.IFO copy all its titles to our IFO
			if(iVTS == VTS/* && iTT == TT*/ && iTT <= TT)
			{
				CopyMemory(&IFOBuffer[pTT_SRPT + 8 + (iTT - 1) * 12], &orgIFOBuffer[pTT], 6);
				IFOBuffer[pTT_SRPT + 8 + (iTT - 1) * 12 + 0x0] &= 0xFC; // remove prohibited user ops
				byteCastW((pTT_SRPT + 8 + (iTT - 1) * 12 + 0x6), 1);
				byteCastW((pTT_SRPT + 8 + (iTT - 1) * 12 + 0x7), iTT);
				intCastW((pTT_SRPT + 8 + (iTT - 1) * 12 + 0x8), 6);
			}
		}

		// add the additional entries into the header of TT_SRPT
		sintCastW(pTT_SRPT, TT);
		intCastW(pTT_SRPT + 0x4, 12 * TT + 8);

		// modify the jump to title command in the first play programm chain
		int FP_PGC = intCast(0x0084);
		int offCmd = sintCast(FP_PGC + 0x00E4) + FP_PGC;
        byteCastW(offCmd + 0x0008 + 5, TT);
	}
}

// set end of title set
void IFO::SetLastBUPSector(int sector)
{
	lastBUPSector = sector;
}

// set begining of the menu VOB
void IFO::SetFirstMenuVOBSector(int sector)
{
	MenuVOBStartSECT = sector;
}

// set beginning of the title VOB
void IFO::SetFirstTitleVOBSector(int sector)
{
	TitleVOBStartSECT = sector;
}

// set start of earch title set (first sector of each IFO)
int IFO::VMG_SetTitleStartSector(int title, int sector)
{
	for(int i = 0; i < titleStartLUTEntries; i++)
	{
		// if the specified title set is found set its new sector
		if(titleStartLUT[i].userData == title)
		{
			titleStartLUT[i].targetSector = sector;
		}
	}
	
	return TRUE;
}

// *** sector LUT entry manipulation routines ***
// returns the next entry in the sector LUT
int IFO::GetNextSectorEntry(int reference)
{
	// if last entry was processed or there are no entries then return maximum sector number
	int sectorEntry;

	switch(reference)
	{
		case REFERENCE_MENU_VOB:  { { if(currLUTVOBEntry == -1) return INT_MAX; } sectorEntry = sectorLUT[currLUTVOBEntry].srcSector; } break;
		case REFERENCE_TITLE_VOB: { { if(currLUTVOBEntry == -1) return INT_MAX; } sectorEntry = sectorLUT[currLUTVOBEntry].srcSector; } break;
		case REFERENCE_IFO:		  { { if(currLUTIFOEntry == -1) return INT_MAX; } sectorEntry = sectorLUT[currLUTIFOEntry].srcSector; } break;
		default: _ASSERT(FALSE);
	}

	int minSectorEntry = INT_MAX;
	int minIndex = -1;

	// search for next sector entry (may not be the next in the table)
	for(int i = 0; i < sectorLUTEntries; i++)
	{
		if(sectorLUT[i].srcSector < minSectorEntry && sectorLUT[i].srcSector > sectorEntry && sectorLUT[i].reference == reference)
		{
			minSectorEntry = sectorLUT[i].srcSector;
			minIndex = i;
		}
	}

	switch(reference)
	{
		case REFERENCE_MENU_VOB:  { currLUTVOBEntry = minIndex; } break;
		case REFERENCE_TITLE_VOB: { currLUTVOBEntry = minIndex; } break;
		case REFERENCE_IFO:		  { currLUTIFOEntry = minIndex; } break;
		default: _ASSERT(FALSE);
	}

	return sectorEntry;
}

// sets the translated source sector in the corresponding sector LUT entries
void IFO::SetSectorEntry(int srcSector, int targetSector, int reference)
{
	int i = 0;

	_ASSERT(srcSector >= targetSector);
//	_ASSERT(srcSector == targetSector);
	for(i = 0; i < sectorLUTEntries; i++)
	{
		if(sectorLUT[i].srcSector == srcSector && sectorLUT[i].reference == reference) 
		{
			sectorLUT[i].targetSector = targetSector;
		}
	}
}

// initializes the get-next-sector-entry iterator
void IFO::SetCurrSectorEntry(int srcSector, int reference)
{
	int minSectorEntry = INT_MAX;
	int minIndex = -1;

	// search for next sector entry (may not be the next in the table)
	for(int i = 0; i < sectorLUTEntries; i++)
	{
		if(sectorLUT[i].srcSector < minSectorEntry && sectorLUT[i].srcSector >= srcSector && sectorLUT[i].reference == reference)
		{
			minSectorEntry = sectorLUT[i].srcSector;
			minIndex = i;
		}
	}

	// assume that the new sector has not yet been transferred
	_ASSERT((minIndex < 0) || sectorLUT[minIndex].targetSector == -1);

	switch(reference)
	{
		case REFERENCE_MENU_VOB:  { currLUTVOBEntry = minIndex; } break;
		case REFERENCE_TITLE_VOB: { currLUTVOBEntry = minIndex; } break;
		case REFERENCE_IFO:		  { currLUTIFOEntry = minIndex; } break;
		default: _ASSERT(FALSE);
	}
}

int IFO::GetNumVTS()
{
	if(isVMG != 1) return -1;

	// if real titleset is specified
	int pVMG_VTS_ATRT = intCast(0x00D0) * 2048;

	// process TT_SRPT (title set beginnings)
	if(pVMG_VTS_ATRT != 0)
	{
		return sintCast(pVMG_VTS_ATRT) + 1;
	}

	return 1;
}

int IFO::GetNumVTSLang(int VTSIndex)
{
	if(isVMG != 1 || VTSIndex < 0) return -1;

	int numLang = 0;

	// if real titleset is specified
	if(VTSIndex > 0)
	{
		int pVTS_IFO = GetVTS_IFOPointer(VTSIndex - 1);
		if(pVTS_IFO != -1)
		{
			numLang += sintCast(pVTS_IFO + (0x102 - 0x100));
			numLang += sintCast(pVTS_IFO + (0x202 - 0x100));
		}
	}
	// if menu is specified
	else if(VTSIndex == 0)
	{
			numLang += sintCast(0x102);
	}

	return numLang;
}

int IFO::GetVTSLangDesc(int VTSIndex, int langIndex, char *buffer, int szBuffer)
{
	if(isVMG != 1 || VTSIndex < 0) return -1;

	int pLANG_ENTRY = -1;
	int numLang = 0;

	int menuItem = -1;

	// if real titleset is specified
	if(VTSIndex > 0)
	{
		int pVTS_IFO = GetVTS_IFOPointer(VTSIndex - 1);
		if(pVTS_IFO != -1)
		{
			menuItem = 1;
			numLang += sintCast(pVTS_IFO + (0x102 - 0x100));
			if(numLang > langIndex)
			{
				pLANG_ENTRY = pVTS_IFO + (0x104 + 8 * langIndex - 0x100);
			}
			else
			{
				menuItem = 0;
				int snumLang = numLang;
				numLang += sintCast(pVTS_IFO + (0x202 - 0x100));
				if(numLang > langIndex)
				{
					pLANG_ENTRY = pVTS_IFO + (0x204 + 8 * (langIndex - snumLang) - 0x100);
				}
			}
		}
	}
	// if menu is specified
	else if(VTSIndex == 0)
	{
			numLang += sintCast(0x102);
			if(numLang > langIndex)
			{
				pLANG_ENTRY = 0x104 + 8 * langIndex;
			}
	}

	// read+describe language
	char language[32];
	if(((byteCast(pLANG_ENTRY + 0) >> 2) & 3) == 1)
	{
		sprintf(language, "%c%c", byteCast(pLANG_ENTRY + 2), byteCast(pLANG_ENTRY + 3));
	}
	else
	{
		strcpy(language, "Unspecified Language");
	}

	// read+describe compression format
	char format[16];
	switch((byteCast(pLANG_ENTRY + 0) >> 5) & 7)
	{
		case 0: strcpy(format, " AC3");		  break;
		case 2: strcpy(format, " MPEG-1");	  break;
		case 3: strcpy(format, " MPEG-2EXT"); break;
		case 4: strcpy(format, " LPCM");	  break;
		case 6: strcpy(format, " DTS");		  break;
		case 7: strcpy(format, " SDDS");	  break;
		default: strcpy(format, "");		  break;
	}

	// read+describe sample frequency
	char freq[16];
	switch((byteCast(pLANG_ENTRY + 1) >> 4) & 3)
	{
		case 0: strcpy(freq, " (48kHz)");	 break;
		case 1: strcpy(freq, " (96kHz)");	 break;
		default: strcpy(freq, "");			 break;
	}

	// read+describe number of channels
	char channels[16];
	int num = byteCast(pLANG_ENTRY + 1) & 7;
	sprintf(channels, " %i channels", (num + 1));

	// read+describe number of channels
	char quant[16];
	switch((byteCast(pLANG_ENTRY + 1) >> 6) & 3)
	{
		case 0: strcpy(quant, " 16bit");	 break;
		case 1: strcpy(quant, " 20bit");	 break;
		case 2: strcpy(quant, " 24bit");	 break;
		case 3: strcpy(quant, " DRC");		 break;
		default: _ASSERT(FALSE);		     break;
	}

	// read+describe application mode
	char application[16];
	switch(byteCast(pLANG_ENTRY + 0) & 3)
	{
		case 1: strcpy(application, " karaoke");	 break;
		case 2: strcpy(application, " surround");	 break;
		default: strcpy(application, "");			 break;
	}

	// read+describe application mode
	char extension[64];
	switch(byteCast(pLANG_ENTRY + 5) & 7)
	{
		case 2: strcpy(extension, " for visually impaired");		 break;
		case 3: strcpy(extension, " director's comments");			 break;
		case 4: strcpy(extension, " alternate director's comments"); break;
		default: strcpy(extension, "");								 break;
	}

	int len = strlen(language) + strlen(format) + strlen(freq) + strlen(channels) + strlen(quant) + strlen(application) + strlen(extension);

	// if buffer is too small
	if(szBuffer < len + 4) return -2;

	// write description
	if(len - strlen(language) > 0)
	{
		sprintf(buffer, "%s -%s%s%s%s%s%s", language, format, freq, channels, quant, application, extension);
	}
	else
	{
		sprintf(buffer, "%s", language);
	}

	if(menuItem == 1)		strcat(buffer, " (menu)");
	else if(menuItem == 0)  strcat(buffer, " (title set)");

	return len;
}

int IFO::IsVTSLangMenuItem(int VTSIndex, int langIndex)
{
	if(isVMG != 1 || VTSIndex < 0) return -1;

	int pLANG_ENTRY = -1;
	int numLang = 0;

	int menuItem = -1;

	// if real titleset is specified
	if(VTSIndex > 0)
	{
		int pVTS_IFO = GetVTS_IFOPointer(VTSIndex - 1);
		if(pVTS_IFO != -1)
		{
			menuItem = 1;
			numLang += sintCast(pVTS_IFO + (0x102 - 0x100));
			if(numLang <= langIndex)
			{
				menuItem = 0;
			}
		}
	}
	else
	{
		menuItem = 1;
	}

	return menuItem;
}

int IFO::GetNumVTSSub(int VTSIndex)
{
	if(isVMG != 1 || VTSIndex < 0) return -1;

	int numSub = 0;

	// if real titleset is specified
	if(VTSIndex > 0)
	{
		int pVTS_IFO = GetVTS_IFOPointer(VTSIndex - 1);
		if(pVTS_IFO != -1)
		{
			numSub += sintCast(pVTS_IFO + (0x154 - 0x100));
			numSub += sintCast(pVTS_IFO + (0x254 - 0x100));
		}
	}
	// if menu is specified
	else if(VTSIndex == 0)
	{
			numSub += sintCast(0x154);
	}

	return numSub;
}

int IFO::GetVTSSubDesc(int VTSIndex, int subIndex, char *buffer, int szBuffer)
{
	if(isVMG != 1 || VTSIndex < 0) return -1;

	int pSUB_ENTRY = -1;
	int numSub = 0;

	int menuItem = -1;

	// if real titleset is specified
	if(VTSIndex > 0)
	{
		int pVTS_IFO = GetVTS_IFOPointer(VTSIndex - 1);
		if(pVTS_IFO != -1)
		{
			menuItem = 1;
			numSub += sintCast(pVTS_IFO + (0x154 - 0x100));
			if(numSub > subIndex)
			{
				pSUB_ENTRY = pVTS_IFO + (0x156 + 6 * subIndex - 0x100);
			}
			else
			{
				menuItem = 0;
				int snumSub = numSub;
				numSub += sintCast(pVTS_IFO + (0x254 - 0x100));
				if(numSub > subIndex)
				{
					pSUB_ENTRY = pVTS_IFO + (0x256 + 6 * (subIndex - snumSub) - 0x100);
				}
			}
		}
	}
	// if menu is specified
	else if(VTSIndex == 0)
	{
			numSub += sintCast(0x154);
			if(numSub > subIndex)
			{
				pSUB_ENTRY = 0x156 + 6 * subIndex;
			}
	}

	// read+describe language
	char language[32];
	if((byteCast(pSUB_ENTRY + 0) & 3) == 1)
	{
		sprintf(language, "%c%c", byteCast(pSUB_ENTRY + 2), byteCast(pSUB_ENTRY + 3));
	}
	else
	{
		sprintf(language, "Unspecified Language");
	}

	int len = strlen(language);

	// if buffer is too small
	if(szBuffer < len + 1) return -2;

	// write description
	strcpy(buffer, language);
	if(menuItem == 1)	   strcat(buffer, " (menu)");
	else if(menuItem == 0) strcat(buffer, " (title set)");

	return len;
}

int IFO::IsVTSSubMenuItem(int VTSIndex, int subIndex)
{
	if(isVMG != 1 || VTSIndex < 0) return -1;

	int pSUB_ENTRY = -1;
	int numSub = 0;

	int menuItem = -1;

	// if real titleset is specified
	if(VTSIndex > 0)
	{
		int pVTS_IFO = GetVTS_IFOPointer(VTSIndex - 1);
		if(pVTS_IFO != -1)
		{
			menuItem = 1;
			numSub += sintCast(pVTS_IFO + (0x154 - 0x100));
			if(numSub <= subIndex)
			{
				menuItem = 0;
			}
		}
	}
	else
	{
		menuItem = 1;
	}

	return menuItem;
}

int IFO::GetVTS_IFOPointer(int VTSIndex)
{
	int pVTS_IFO = -1;

	// get table of vts stream specifications (copy of vts ifos)
	int pVMG_VTS_ATRT = intCast(0x00D0) * 2048;

	// process TT_SRPT (title set beginnings)
	if(pVMG_VTS_ATRT != 0)
	{
		int numTitles = sintCast(pVMG_VTS_ATRT);
		_ASSERT(numTitles < 100);
		int lastByteVMG_VTS_ATRT = intCast(pVMG_VTS_ATRT + 4) + pVMG_VTS_ATRT;
		// pointer to one entry
		int pVTS_ATRT = intCast(pVMG_VTS_ATRT + 8 + VTSIndex * 4) + pVMG_VTS_ATRT;
		_ASSERT(pVTS_ATRT < lastByteVMG_VTS_ATRT);
		_ASSERT((pVTS_ATRT + intCast(pVTS_ATRT)) <= lastByteVMG_VTS_ATRT);
		// pointer to copy of ifo
		pVTS_IFO = pVTS_ATRT + 8;
	}

	return pVTS_IFO;
}

int IFO::GetNumSub(int reference)
{
	// return number
	if(reference == REFERENCE_MENU_VOB) return sintCast(0x154);
	else		 					    return sintCast(0x254);

	return 0;
}

int IFO::GetNumLang(int reference)
{
	// return number
	if(reference == REFERENCE_MENU_VOB) return sintCast(0x102);
	else							    return sintCast(0x202);

	return 0;
}

int IFO::GetLangCompression(int reference, int langIndex)
{
	int pLANG_ENTRY = -1;

	// retrieve pointer
	if(reference == REFERENCE_MENU_VOB) pLANG_ENTRY = 0x104 + 8 * langIndex;
	else							    pLANG_ENTRY = 0x204 + 8 * langIndex;

	// read+describe compression format
	int type = (byteCast(pLANG_ENTRY + 0) >> 5) & 7;

	return type;
}

int IFO::GetPGCMarks(int reference, int **markSECT)
{
	int pPGCI;

	if(reference == REFERENCE_MENU_VOB)
	{
		int pPGCI_UT;
		if(!isVMG) pPGCI_UT = intCast(0x00D0) * 2048;
		else	   pPGCI_UT = intCast(0x00C8) * 2048;

		// process VMGM_PGCI_UT (program chains)
		int numLanguageUnits = sintCast(pPGCI_UT);
		int lastBytePGCI_UT = intCast(pPGCI_UT + 4) + pPGCI_UT;
		// process all Language Units
		for(int i = 0; i < numLanguageUnits; i++)
		{
			int poffPGCI = pPGCI_UT + (i * 8 + 4) + 8;
			// check whether language pointer is beyond VMGM_PGCI_UT border
			_ASSERT(poffPGCI < lastBytePGCI_UT);
			pPGCI = intCast(poffPGCI) + pPGCI_UT;
		}
	}
	else
	{
		pPGCI = intCast(0x0CC) * 2048;
	}

	int numProgramChains = sintCast(pPGCI);
	int lastBytePGCI = intCast(pPGCI + 4) + pPGCI;

	if(this->markSECT != NULL) 
	{
		delete AVStreamID;
		delete this->markSECT;
	}
	this->markSECT = new int[numProgramChains];
	this->AVStreamID = new AVTABLE[numProgramChains];

	ZeroMemory(AVStreamID, numProgramChains * sizeof(AVTABLE));
	// process all Program Chains
	for(int j = 0; j < numProgramChains; j++)
	{
		int poffPGC = pPGCI + (j * 8 + 4) + 8;
		// check whether language pointer is beyond VMGM_LU border
		_ASSERT(poffPGC < lastBytePGCI);
		int	pPGC = intCast(poffPGC) + pPGCI;
			
		for(int i = 0; i < GetNumLang(reference); i++)
		{
			AVStreamID[j].audioAvailable[i] = byteCast(pPGC + 0x000C + i * 2) >> 7;
			// if the stream is available
			if(AVStreamID[j].audioAvailable[i])
			{
				AVStreamID[j].audioStreamNum[i] = byteCast(pPGC + 0x000C + i * 2) & 7;
				int type = GetLangCompression(reference, i);
				// this is a mpeg audio stream
				if(type == 2 || type == 3) 
				{
					AVStreamID[j].audioStreamNum[i] |= 0xC0;
					AVStreamID[j].audioStreamType[i] = STREAM_AUDIO;
				}
				else
				{
					switch(type)
					{
						case 0: 
						case 7: AVStreamID[j].audioStreamNum[i] += 0x80; break;
						case 6: AVStreamID[j].audioStreamNum[i] += 0x88; break;
						case 4: AVStreamID[j].audioStreamNum[i] += 0xA0; break;
						default: _ASSERT(FALSE);
					}
					AVStreamID[j].audioStreamType[i] = STREAM_SUBSTREAM;
				}
			}
		}

		for(i = 0; i < GetNumSub(reference); i++)
		{
			AVStreamID[j].videoAvailable[i] = byteCast(pPGC + 0x001C + i * 4) >> 7;
			// if the stream is available
			if(AVStreamID[j].videoAvailable[i])
			{
				int maxAVStreamID = -1;
				for(int k = 0; k < 4; k++)
				{
					AVStreamID[j].videoStreamNum[i * 4 + k] = (byteCast(pPGC + 0x001C + i * 4 + k) & 31) + 0x20;
					// ** workaround ** because some entries may be 0, save maximum stream id
					if(AVStreamID[j].videoStreamNum[i * 4 + k] < maxAVStreamID)
					{
						maxAVStreamID = AVStreamID[j].videoStreamNum[i * 4 + k];
					}
				}
				// ** workaround ** if for non 0 substream number there is a 0 entry, then replace it with a non 0 entry (maxAVStreamID) if available
				if(i != 0)
				{
					for(int k = 0; k < 4; k++)
					{
						if(AVStreamID[j].videoStreamNum[i * 4 + k] == 0) AVStreamID[j].videoStreamNum[i * 4 + k] = maxAVStreamID;
					}
				}
			}
		}

		int numCells = byteCast(pPGC + 0x0003);
		int	pcellInfo = sintCast(pPGC + 0x00E8) + pPGC;
		// process first cell (to be complete all cells should be taken into account with holes in the middle of the programm)
		this->markSECT[j] = intCast(pcellInfo + 0x0008);
	}

	// sort programm chains in ascending order (according to first sector beginning cell)
	for(j = 0; j < numProgramChains; j++)
	{
		AVTABLE tTable;
		int tMark = INT_MAX;
		int index = -1;
		// search for next PGC
		for(int i = j; i < numProgramChains; i++)
		{
			if(this->markSECT[i] < tMark) 
			{
				tMark = this->markSECT[i];
				index = i;
			}
		}
		// exchange PGCs
		tMark = this->markSECT[j];
		tTable = this->AVStreamID[j];

		this->markSECT[j] = this->markSECT[index];
		this->AVStreamID[j] = this->AVStreamID[index];

		this->markSECT[index] = tMark;
		this->AVStreamID[index] = tTable;
	}

	*markSECT = this->markSECT;

	return numProgramChains;

}

int IFO::GetAVTable(int reference, AVTABLE **AVStreamID)
{
	*AVStreamID = this->AVStreamID;
	return 1;
}

void IFO::SetMovieOnly(bool movieOnly)
{
	this->movieOnly = movieOnly;
}

void IFO::CopyAudioVideoAttributes(unsigned char *srcIFOBuffer)
{
	if(movieOnly)
	{
		int VMG_VTS_ATRT = intCast(0x00D0) * 2048;
		int VTS_ATRT1 = VMG_VTS_ATRT + intCast(VMG_VTS_ATRT + 8);
		CopyMemory(&IFOBuffer[VTS_ATRT1 + 4], &srcIFOBuffer[0x0022], 4);
		CopyMemory(&IFOBuffer[VTS_ATRT1 + 8], &srcIFOBuffer[0x0100], 0x300);
	}
}

unsigned char *IFO::GetIFOBuffer()
{
	return IFOBuffer;
}