/*****************************************************************************
 * MacroBlock.h: Read and Write Macroblock Class
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#pragma once

#include "ISO_DEFS.h"

#include "VLC.h"
#include "Block.h"

// structure for VLC encoding of macroblock type
typedef struct tagMBTYPE
{
	char macroblock_quant;
	char macroblock_motion_forward;
	char macroblock_motion_backward;
	char macroblock_pattern;
	char macroblock_intra;
} MBTYPE;

#define MV_FORWARD	0
#define MV_BACKWARD	1

#define MACROBLOCK_EMPTY   0x80000000
#define MACROBLOCK_INVALID 0x40000000
#define MACROBLOCK_QUANT   0x20000000

class MacroBlock
{
public:
	MacroBlock(void);
	~MacroBlock(void);

	virtual int AttachBuffer(unsigned char *buffer, int length, int bit_startpos);
	virtual int WriteDataBuffer(unsigned char *buffer, int bit_startpos, int irreductible);

	virtual void SetDCYUVpred(int *YUVpred); // sets the YUV predictors
	virtual void GetDCYUVpred(int *YUVpred); // gets the YUV predictors
	virtual void GetDCCoef(int *DCCoef);	 // gets predicted values

	virtual void AddSkippedCount(int skippedCnt);			// adds to the macroblock address increment 
															// (skipped macroblocks) the given value
	virtual int  GetSkippedCount();							// returns the macroblock address increment
															// the number of (skipped macroblocks)
	virtual int  IsIntra();									// returns whether the macroblock is intra coded
	virtual int  IsCodedQuantizerVal();						// returns whether the macroblock has coded its own quantization value
	virtual int  SetCodedQuantizer();						// configures the block so that a quantizer is coded

	// parameters that need to be set before buffer attachment
	int pictureType;
	int intraDCPrecision;
	int picture_structure;
	int frame_pred_frame_dct;
	int concealment_motion_vectors;
	int qScaleType;
	int intraVLCFormat;
	int alternateScan;
	int chroma_format;
	int f_code[2][2];
	int quantizerScaleCode;
	int absNumPicture;
	int absNumMacroBlock;
	int absNumBlock;
	int demoMode;

	// buffer
	short          *bl_coefBuffer;
	unsigned char  *bl_runBuffer;
	int			    bl_numCoef; // number of coefficients the block has decoded

protected:
	unsigned char *pBuffer;
	int bufferSize;

	static const MBTYPE IFrame[2];				// I-Frame Macroblock-Type Table - ISO 13818-2 Table B-2 (settings part)
	static const MBTYPE PFrame[7];				// P-Frame Macroblock-Type Table - ISO 13818-2 Table B-3 (settings part)
	static const MBTYPE BFrame[11];				// B-Frame Macroblock-Type Table - ISO 13818-2 Table B-4 (settings part)

	static const VLC_ENTRY IFrameVLC[2];		// I-Frame Macroblock-Type Table - ISO 13818-2 Table B-2 (VLC part)
	static const VLC_ENTRY PFrameVLC[7];		// P-Frame Macroblock-Type Table - ISO 13818-2 Table B-3 (VLC part)
	static const VLC_ENTRY BFrameVLC[11];		// B-Frame Macroblock-Type Table - ISO 13818-2 Table B-4 (VLC part)

	static const char IFrameIDX[2];				// I-Frame Macroblock-Type Table Indices - ISO 13818-2 Table B-2 
	static const char PFrameIDX[6];				// P-Frame Macroblock-Type Table Indices - ISO 13818-2 Table B-3 
	static const char BFrameIDX[6];				// B-Frame Macroblock-Type Table Indices - ISO 13818-2 Table B-4 

	static const VLC_ENTRY MBAddressVLC[34];	// Macroblock Addressing Table - ISO 13818-2 Table B-1 (VLC part)
	static const VLC_ENTRY MBPatternVLC[64];	// Macroblock Pattern Table    - ISO 13818-2 Table B-9 (VLC part)
	static const char MBPattern[64];			// Macroblock Pattern Table    - ISO 13818-2 Table B-9 (settings part)
	static const char MBPatternINV[64];			// Inverse Macroblock Pattern Table - Inverse ISO 13818-2 Table B-9 (settings part)

	static const VLC_ENTRY MVCodeVLC[17];		// Motion Vector Code Table    - ISO 13818-2 Table B-10 (VLC part)

	static const char MBAddressIDX[8];			// Macroblock Addressing Table Indices - ISO 13818-2 Table B-1 
	static const char MBPatternIDX[9];			// Macroblock Pattern Table Indices	   - ISO 13818-2 Table B-9 

	static const char MVCodeIDX[7];				// Motion Vector Code Table Indices    - ISO 13818-2 Table B-10

	Block block[12];							// DCT Blocks

	char pattern_code[12];						// block select flags
	char test_pattern_code[12];					// block select flags

	int DCCoef[12];								// DC coefficients for I-Pictures
	int YUVpred[3];

	static VLC bstream;
	static VLC tBitStream;

	int MBAddressInc, MBAddressIncSize;
	int MBTypeIdx, MBTypeSize;
	char dct_type, dct_typeSize;

	// macroblock type classification
	MBTYPE *MBType;

	// store for data before quantizer scale
	int preQScaleStore[16];
	int preQScaleStoreBitSize;

	// store for data after quantizer scale
	int postQScaleStore[16];
	int postQScaleStoreBitSize;

	// relevant parameters

	int ReadMotionData(int motion_vector_count, int mv_format, int dmv, int mv_dir);
	int ReadMotionVector(int dmv, int mv_dir, int num);
	int ReadCodedBlockPattern();
	int WriteCodedBlockPattern();

	int orgQuantizerScaleCode;

	// debug purposes only
	int testBitSize;

	virtual int ValidateMacroBlock(int irreductible);
};
